/*
 * mpgedit file playback API
 *
 * Copyright (C) 2001-2003 Adam Bernstein
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307  USA.
 */


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/stat.h>
#include "portability.h"
#include "header.h"
#include "mp3time.h"
#include "mpegindx.h"
#include "xing_header.h"
#include "mpegfio.h"
#include "mp3_header.h"
#include "mpegcurses.h"
#include "decoder.h"
#include "plugin.h"
#include "version.h"
#include "playif.h"
#include "p_playif.h"
#include "editif.h"

static plugin_fcn_dsc *fp_mpgdecoder;
static int            fp_mpgdecoder_refcnt;


#define MPGEDIT_PLAYCTX(a) ((playctx *) (a))

/*
 * Obviously, all of these static functions need to be moved into mpegindx.c
 * Major refactoring here!!!
 */
void *mpgedit_play_init(char *name)
{
    char    *loaderrmsg;
    playctx *ctx = NULL;
    long    rsts = 0;
    long    sec;
    long    usec;
    FILE    *fp = NULL;
    struct stat sbuf;

    if (name) {
        if (stat(name, &sbuf) == -1) {
            return NULL;
        }
    
        if (!S_ISREG(sbuf.st_mode)) {
            return NULL;
        }
    
        fp = fopen(name, "rb");
        if (!fp) {
            fprintf(stderr, "fopen: test1.mp3 failed %s\n", strerror(errno));
            return NULL;
        }
    }

    ctx = calloc(1, sizeof(playctx));
    if (!ctx) {
        rsts = -1;
        goto clean_exit;
    }
    ctx->name   = name;
    ctx->mpegfp = fp;
    mpeg_file_stats_init(&ctx->stats, name, 0,0);


    fp_mpgdecoder = mpgfio_alloc_decoder_plugin();
    if (!fp_mpgdecoder) {
        fprintf(stderr, "Failed allocating decoder plugin\n");
        rsts = -1;
        goto clean_exit;
    }
    fp_mpgdecoder_refcnt++;

    /*
     * Load MP3 file playback plugin.
     */
    loaderrmsg = mpegfio_load_decoder_plugin("libmpgedit_decoder",
                                             fp_mpgdecoder);
    if (loaderrmsg) {
        fprintf(stderr,
                "Error while loading plugin 'libmpgedit_decoder': %s\n",
                loaderrmsg);
        fprintf(stderr, "%s\n",
            "Warning: decoder.so plugin not found; default decoder used");
        fprintf(stderr, "%s\n",
            "         set LD_LIBRARY_PATH or install with ldconfig\n");
        free(loaderrmsg);
        rsts = -1;
        goto clean_exit;
    }
    ctx->mpctx = fp_mpgdecoder_alloc();
    if (!ctx->mpctx) {
        fprintf(stderr, "allocating mpgdecoder context failed\n");
        rsts = -1;
        goto clean_exit;
    }
    ctx->playfp = fp_mpgdecoder_open(ctx->mpctx, 44100, 0);
    if (!ctx->playfp) {
#if 0
        fprintf(stderr, "Unable to open playback plugin");
#endif
        rsts = -1;
        goto clean_exit;
    }
    fp_mpgdecoder_init(ctx->mpctx);

    ctx->fsize = ctx->sec = ctx->usec = -1;
    if (name) {
        ctx->indxfp = init_timeindex_file(ctx->mpegfp, ctx->name, 
                                          &ctx->indx_name);
        if (ctx->indxfp) {
            /*
             * Get file size from index file
             */
            rsts = mp3edit_get_size_from_index(ctx->indxfp, &sec, &usec);
            if (rsts != -1) {
                ctx->fsize = rsts;
                ctx->sec   = sec;
                ctx->usec  = usec;
            }
        }
        memset(&ctx->header, 0,  sizeof(ctx->header));
        memset(&ctx->xingh,  0,  sizeof(ctx->xingh));
        ctx->has_xing = mpegfio_has_xing_header(ctx->mpegfp,
                                                &ctx->header,
                                                &ctx->xingh, 0);
    
        mpeg_file_stats_init(&ctx->stats, name, ctx->has_xing, 0);
    }

clean_exit:
    if (rsts == -1) {
        if (fp) {
            fclose(fp);
        }
        if (ctx) {
            if (ctx->playfp) {
                fp_mpgdecoder_close(ctx->mpctx);
            }
            if (ctx->mpctx) {
                fp_mpgdecoder_free(ctx->mpctx);
            }
            free(ctx);
            ctx = NULL;
        }
        /*
         * There may be another instance of a play context that will be using
         * this global pointer to the plugin function pointers.  Only free
         * this when the reference count goes to zero.
         */
        fp_mpgdecoder_refcnt--;
        if (fp_mpgdecoder == 0 && fp_mpgdecoder) {
            free(fp_mpgdecoder);
        }
    }

    return ctx;
}


int mpgedit_play_seek_time(void *_ctx, long sec, long msec)
{
    playctx          *ctx = MPGEDIT_PLAYCTX(_ctx);
    long             offset;

    if (!ctx) {
        return 0;
    }
    memset(&ctx->mpegtval, 0, sizeof(ctx->mpegtval));
    memset(&ctx->stime,    0, sizeof(ctx->stime));

    ctx->stime.units = sec;
    ctx->stime.usec  = msec * 1000;
    
    offset = mpeg_time_seek_starttime(
                 ctx->mpegfp, 
                 ctx->indxfp, 
                 &ctx->stime, 
                 &ctx->mpegtval, 
                 ctx->has_xing);
    return offset;
}



void mpgedit_play_set_status_callback(void *_ctx,
                                      int (*status)(void *, long, long),
                                      void *data)
{
    playctx *ctx = MPGEDIT_PLAYCTX(_ctx);

    if (!ctx) {
        return;
    }
    ctx->fp_status = status;
    ctx->dp_status = data;
}


int mpgedit_play_frame(void *_ctx)
{
    playctx          *ctx = MPGEDIT_PLAYCTX(_ctx);
    int              eof;
    int              found;
    long             sec;
    long             usec;

    if (!ctx) {
        return 0;
    }
    found = read_next_mpeg_frame(ctx->mpegfp, &ctx->mpegiobuf,
                                 &ctx->header, &eof);
    if (found) {

        /*
         * Add frame time to running total. When the current time, tnow,
         * is advanced 1 second, call the notification callback.
         */
        mpeg_time_frame_increment(&ctx->stime, &ctx->header);
        mpeg_file_stats_gather(&ctx->stats, &ctx->header);
        mpeg_time_gettime(&ctx->stime, &sec, &usec);
        ctx->cursec  = sec;
        ctx->curusec = usec;
#if 1
        /*
         * adam/TBD:
         * These two lines imply these variables are equivalent.
         * Is this generally true throughout editif and playif??
         */
        ctx->stats_frame_header = ctx->header;
        ctx->mpegtval           = ctx->stime;
#endif


        fp_mpgdecoder_play_frame(
            ctx->mpctx, (void *) ctx->playfp,
            mpeg_file_iobuf_getptr(&ctx->mpegiobuf),
            ctx->header.frame_size, ctx->header.sample_rate,
            ctx->header.channel_mode != 3);

        mpeg_file_iobuf_setstart(&ctx->mpegiobuf,
                                 mpeg_file_iobuf_getstart(&ctx->mpegiobuf) +
                                 ctx->header.frame_size);
        if (ctx->fp_status) {
            found = ctx->fp_status(ctx->dp_status, sec, usec/1000);
        }

    }
    if (eof) {
        mpeg_file_iobuf_clear(&ctx->mpegiobuf);
    }
    return found;
}


void mpgedit_play_close(void *_ctx)
{
    playctx *ctx = MPGEDIT_PLAYCTX(_ctx);

    if (!ctx) {
        return;
    }
    if (ctx->mpegfp) {
        fclose(ctx->mpegfp);
    }
    if (ctx->indxfp) {
        fclose(ctx->indxfp);
    }
    fp_mpgdecoder_close(ctx->mpctx);
    memset(ctx, 0, sizeof(*ctx));
    free(ctx);
}


long mpgedit_play_total_size(void *_ctx)
{
    return _ctx ? MPGEDIT_PLAYCTX(_ctx)->fsize : 0;
}


long mpgedit_play_total_sec(void *_ctx)
{
    return _ctx ? MPGEDIT_PLAYCTX(_ctx)->sec : 0;
}


long mpgedit_play_total_msec(void *_ctx)
{
    if (!_ctx) {
        return -1;
    }
    if (MPGEDIT_PLAYCTX(_ctx)->usec == -1) {
        return -1;
    }
    return MPGEDIT_PLAYCTX(_ctx)->usec/1000;
}


long mpgedit_play_current_sec(void *_ctx)
{
    return _ctx ? MPGEDIT_PLAYCTX(_ctx)->cursec : 0;
}


long mpgedit_play_current_msec(void *_ctx)
{
    if (!_ctx) {
        return -1;
    }
    if (MPGEDIT_PLAYCTX(_ctx)->curusec == -1) {
        return -1;
    }
    return MPGEDIT_PLAYCTX(_ctx)->curusec/1000;
}


#if defined(UNIT_TEST)
int main(void)
{
    int go;
    void *ctx;
    int again;

    do {
        ctx = mpgedit_play_init("test1.mp3");
        if (!ctx) {
            return 1;
        }
    } while (again);

    do {
        go = mpgedit_play_frame(ctx);
    } while (go);

    return 0;
}
#endif
