/*
 * xmpgedit volume control interface
 *
 * Copyright (C) 2003 Adam Bernstein
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307  USA.
 */

#include "portability.h"
#include <fcntl.h>
#include "volumeif.h"

typedef struct _volume_control_handle {
    int          fd;     /* UNIX fd/Win32 handle to mixer device */
    unsigned int id;     /* Win32 volume control ID              */
} volume_control_handle;


#if defined(WIN32)
/*
 * Win32 implementation.  Tested on NT4 and W2K.
 */
#include "wmixer.h"

void mpgedit_volume_set(void *fd, int lvol, int rvol)
{
    volume_control_handle *h = (volume_control_handle *) fd;
    /* Somehow use lvol and rvol both */

    /* Multiply by 655 to get maximum volume for 100 */
    lvol *= WIN32_MIXER_SCALE_FACTOR;
    set_control_details(h->fd, h->id, lvol);
}


int mpgedit_volume_get(void *fd, int *lvol, int *rvol)
{
    int                   volume;
    volume_control_handle *h = (volume_control_handle *) fd;

    volume = get_control_details(h->fd, h->id);
    if (volume != -1) {
        *lvol = *rvol = volume / WIN32_MIXER_SCALE_FACTOR;
    }
    return volume;
}


void *mpgedit_volume_open(void)
{
    volume_control_handle *h;

    h = (volume_control_handle *) calloc(1, sizeof(volume_control_handle));
    if (!h) {
        return NULL;
    }
    h->id = open_mixer_pcm_volume_control(&h->fd);

    return h;
}


void mpgedit_volume_close(void *ctx)
{
    volume_control_handle *h;

    if (!ctx) {
        return;
    }

    h = (volume_control_handle *) ctx;
    close_mixer_control(h->fd);
    free(h);
}


#elif defined(__linux)
/*
 * OSS implementation.  Only tested on Linux.
 */
#include <sys/soundcard.h>
#include <sys/ioctl.h>


void mpgedit_volume_set(void *fd, int lvol, int rvol)
{
    int                   mixer_val;
    volume_control_handle *h = (volume_control_handle *) fd;

    if (!fd) {
        return;
    }
    lvol = (lvol < 0)   ? 0     : lvol;
    lvol = (lvol > 100) ? 100   : lvol;
    rvol = (rvol < 0)   ? 0     : rvol;
    rvol = (rvol > 100) ? 100   : rvol;
    if (h->fd != -1) {
        mixer_val = ((rvol & 0xff) << 8) | (lvol & 0xff);
        ioctl(h->fd, SOUND_MIXER_WRITE_PCM, &mixer_val);
    }
}

int mpgedit_volume_get(void *fd, int *lvol, int *rvol)
{
    int mixer_val;
    volume_control_handle *h = (volume_control_handle *) fd;

    if (!h || h->fd == -1) {
        return -1;
    }

    lseek(h->fd, 0, SEEK_SET);
    if (ioctl(h->fd, SOUND_MIXER_READ_PCM, &mixer_val) == -1) {
        return -1;
    }
    lseek(h->fd, 0, SEEK_SET);

    *lvol = mixer_val & 0xff;
    *rvol = (mixer_val >> 8) & 0xff;
    return 0;
}


void *mpgedit_volume_open(void)
{
    volume_control_handle *h;

    h = (volume_control_handle *) calloc(1, sizeof(volume_control_handle));
    if (!h) {
        return NULL;
    }
    h->fd = open(MIXER_DEVICE, O_RDWR);

    return h;
}


void mpgedit_volume_close(void *ctx)
{
    volume_control_handle *h;

    if (!ctx) {
        return;
    }
    h = (volume_control_handle *) ctx;
    if (h->fd != -1) {
        close(h->fd);
    }
    free(h);
}


#else

void mpgedit_volume_set(void *fd, int lvol, int rvol)
{
}


int mpgedit_volume_get(void *fd, int *lvol, int *rvol)
{
    *lvol = *rvol = 0;
    return -1;
}



void *mpgedit_volume_open(void)
{
    return NULL;
}



void mpgedit_volume_close(void *ctx)
{
}



#endif
